# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful functions for testing."""

from typing import Optional
import numpy as np
import pyspiel


def random_playout(state: pyspiel.State, seed: Optional[int] = None):
  """Plays random actions until the state is terminal."""
  rng = np.random.RandomState(seed)
  while not state.is_terminal():
    state.apply_action(rng.choice(state.legal_actions()))
  return state
